==> tcfs-2.2.0a.tar.gz: linux/fs/tcfs/file.c <==
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/stat.h>
#include <linux/mm.h>
#include <linux/tcfs_fs.h>
#include <linux/malloc.h>
#include <linux/pagemap.h>

#include <asm/segment.h>
#include <asm/system.h>

#include "desitfc.h"

/* These are the function needed to handle keys hash table */
#include "hash.h"

static int tcfs_file_mmap(struct inode *, struct file *, struct vm_area_struct *);
static int tcfs_file_read(struct inode *, struct file *, char *, int);
static int tcfs_file_write(struct inode *, struct file *, const char *, int);
static int tcfs_fsync(struct inode *, struct file *);

static struct file_operations tcfs_file_operations = {
	NULL,			/* lseek - default */
	tcfs_file_read,		/* read */
	tcfs_file_write,	/* write */
	NULL,			/* readdir - bad */
	NULL,			/* select - default */
	tcfs_ioctl,		/* ioctl - default */
	tcfs_file_mmap,		/* mmap */
	NULL,			/* no special open is needed */
	NULL,			/* release */
	tcfs_fsync,		/* fsync */
};

struct inode_operations tcfs_file_inode_operations = {
	&tcfs_file_operations,	/* default file operations */
	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	tcfs_readpage,		/* readpage */
	NULL,			/* writepage */
	NULL,			/* bmap */
	NULL			/* truncate */
};

static inline void revalidate_inode(struct tcfs_server * server, struct inode * inode)
{
	struct tcfs_fattr fattr;

	if (jiffies - TCFS_READTIME(inode) < TCFS_ATTRTIMEO(inode)) {
		return;
	}
	TCFS_READTIME(inode) = jiffies;
	if (tcfs_proc_getattr(server, TCFS_FH(inode), &fattr) == 0) {
		tcfs_refresh_inode(inode, &fattr);
		if (fattr.mtime.seconds == TCFS_OLDMTIME(inode)) {
			if ((TCFS_ATTRTIMEO(inode) <<= 1) > server->acregmax)
				TCFS_ATTRTIMEO(inode) = server->acregmax;
			if (S_ISREG(inode->i_mode) && 
				(TCFS_IS_SHARED(inode) || 
				TCFS_IS_SECURE(inode->u.tcfs_i.tcfs_fl.cflag)) && 
				inode->i_size>0 && inode->u.tcfs_i.tcfs_fl.bf.spure>0) {
				inode->i_size-=8-inode->u.tcfs_i.tcfs_fl.bf.spure;
			}
			return;
		}
		TCFS_OLDMTIME(inode) = fattr.mtime.seconds;
	}
	if (tcfs_proc_geteattr(&(inode->i_sb->u.tcfs_sb.x_server),inode->u.tcfs_i.pathname,&(inode->u.tcfs_i.tcfs_fl.cflag))<0) {
#ifdef DEBUG_TCFS
	printk("TCFS: tcfs_proc_geteattr returned error\n");
#endif
		inode->u.tcfs_i.tcfs_fl.bf.mb=0;
		inode->u.tcfs_i.tcfs_fl.bf.spure=0;
		inode->u.tcfs_i.tcfs_fl.bf.unknow=0;
	}
	if (S_ISREG(inode->i_mode) && 
		(TCFS_IS_SHARED(inode) || 
		TCFS_IS_SECURE(inode->u.tcfs_i.tcfs_fl.cflag)) && 
		inode->i_size>0 && inode->u.tcfs_i.tcfs_fl.bf.spure>0) {
		inode->i_size-=8-inode->u.tcfs_i.tcfs_fl.bf.spure;
	}
	invalidate_inode_pages(inode);
}

static int tcfs_file_read(struct inode * inode, struct file * file,
	char * buf, int count)
{
	int result;
	unsigned long old;
	struct hash_entry *htmp;
	struct gid_hash_entry *ghtmp;

#ifdef DEBUG_TCFS
	printk("TCFS: tcfs_file_read %s %d\n",inode->u.tcfs_i.pathname,(int)inode->i_size);
#endif
	if (TCFS_IS_SHARED(inode)) {
		ghtmp=gid_hash_lookup(inode->i_gid);
		if (ghtmp==NULL || ghtmp->ks==NULL) {
			return -EACCES;
		}
	} 

	if (TCFS_IS_SECURE(inode->u.tcfs_i.tcfs_fl.cflag)) {
		htmp=hash_lookup(current->uid);
		if (htmp==NULL || current->uid != inode->i_uid) {
			return -EACCES;
		}
	}
	TCFS_SEMR_DOWN(inode);
	/* DES acts in readpage ... here we do a  normal read ...  */
	revalidate_inode(TCFS_SERVER(inode), inode);
	if (inode==NULL || file==NULL || buf==NULL) {
		TCFS_SEMR_UP(inode);
		return -EIO;
	}
	old=file->f_pos;
	result=generic_file_read(inode, file, buf, count);
/*	if (result>0 && TCFS_IS_SECURE(inode->u.tcfs_i.tcfs_fl.cflag) && 
		inode->u.tcfs_i.tcfs_fl.bf.spure>0 && 
		old+result>inode->i_size - 8+inode->u.tcfs_i.tcfs_fl.bf.spure ) {
		result-=8-inode->u.tcfs_i.tcfs_fl.bf.spure;
	}*/
	TCFS_SEMR_UP(inode);
	return result;
}

static int tcfs_file_mmap(struct inode * inode, struct file * file, struct vm_area_struct * vma)
{
	int ret;
	struct hash_entry *htmp;
	struct gid_hash_entry *ghtmp;

#ifdef DEBUG_TCFS
	printk("TCFS: tcfs_file_mmap %s\n",inode->u.tcfs_i.pathname);
#endif	
	if (TCFS_IS_SHARED(inode)) {
		ghtmp=gid_hash_lookup(inode->i_gid);
		if (ghtmp==NULL || ghtmp->ks==NULL) {
			return -EACCES;
		}
	}
	if (TCFS_IS_SECURE(inode->u.tcfs_i.tcfs_fl.cflag)) {
		htmp=hash_lookup(current->uid);
		if (htmp==NULL) {
			return -EACCES;
		}
	}
	TCFS_SEMR_DOWN(inode);
	revalidate_inode(TCFS_SERVER(inode), inode);
	ret=generic_file_mmap(inode, file, vma);
	TCFS_SEMR_UP(inode);
	return ret;
}

static int tcfs_fsync(struct inode *inode, struct file *file)
{
#ifdef DEBUG_TCFS
	printk("TCFS: tcfs_fsync\n");
#endif
	return 0;
}

extern int tcfs_proc_kwrite(struct inode * inode, int offset,int count, const char *data, struct tcfs_fattr *fattr);

static int tcfs_crywrite(struct inode *inode, struct file *file,const char *buf,
	int count, des_key_schedule *ks)
{
	loff_t newoff,diff;
	struct tcfs_fattr fattr;
	int result,resw;
	char * tmp;
	int towrite=count;
	unsigned char *mbuf=NULL;
	
	while(mbuf==NULL) {
		mbuf=kmalloc(1024,GFP_KERNEL);
		schedule();
	}
	tmp=buf;
	if (file->f_flags & O_APPEND) {
		file->f_pos = inode->i_size;
	}
	do {
		schedule();
		newoff=file->f_pos & 0xfffffc00; /* Calculate offset of DES block */
		diff=file->f_pos-newoff;
		memset(mbuf,'\0',1024);
		result=tcfs_proc_read(&(inode->i_sb->u.tcfs_sb.s_server),TCFS_FH(inode),
					newoff,1024,mbuf,&fattr); /* Read old block */
		if (result <0 ) {/* Error reading file */
			kfree(mbuf);
			return result;
		}
		if (result > 0)
			mkdedes(mbuf,DES8(result),ks);
		if (result==0) { /* End of File */
			if (towrite<1024-diff) { /* Block end into this DES frame */
				int error;
/*				printk("Caso 1 ... subcase 1 %ld\n",(long)newoff);*/
				memcpy_fromfs(&mbuf[diff],tmp,towrite);
				mkendes(mbuf,DES8(towrite+diff),ks);
				resw=tcfs_proc_kwrite(inode,newoff,DES8(towrite+diff),mbuf,&fattr);
				if (resw<0) {
					kfree(mbuf);
					return resw;
				}
				file->f_pos+=towrite;
				tmp+=towrite;
				inode->u.tcfs_i.tcfs_fl.bf.spure=(int)(towrite+diff)%8;
				inode->i_size += towrite;
				towrite=0;
				error=tcfs_proc_seteattr(NULL,&(inode->i_sb->u.tcfs_sb.x_server),inode->u.tcfs_i.pathname,inode->u.tcfs_i.tcfs_fl.cflag,NULL);
				if (error<0) {
					kfree(mbuf);
					return error;
				}
			} else { /* Need another DES frame */
/*				printk("Caso 1 ... subcase 2 %ld\n",(long)newoff);*/
				memcpy_fromfs(&mbuf[diff],tmp,1024);
				mkendes(mbuf,1024,ks);
				resw=tcfs_proc_kwrite(inode,newoff,1024,mbuf,&fattr);
				if (resw<0) {
					kfree(mbuf);
					return resw;
				}
				file->f_pos+=1024-diff;
				towrite-=1024-diff;
				tmp+=1024-diff;
			}
			continue;
		}
		if (result == 1024) { /* File not at the end */
			if (towrite<1024-diff) { /* Block end into this DES frame */
/*				printk("Caso 2 ... subcase 1 %ld\n",(long)newoff);*/
				memcpy_fromfs(&mbuf[diff],tmp,towrite);
				mkendes(mbuf,1024,ks);
				resw=tcfs_proc_kwrite(inode,newoff,1024,mbuf,&fattr);
				if (resw<0) {
					kfree(mbuf);
					return resw;
				}
				tmp+=towrite;
				file->f_pos+=towrite;
				towrite=0;
			} else {
/*				printk("Caso 2 ... subcase 2 %ld\n",(long)newoff);*/
				memcpy_fromfs(&mbuf[diff],tmp,1024-diff);
				mkendes(mbuf,1024,ks);
				resw=tcfs_proc_kwrite(inode,newoff,1024,mbuf,&fattr);
				if (resw<0) {
					kfree(mbuf);
					return resw;
				}
				towrite-=1024-diff;
				tmp+=1024-diff;
				file->f_pos+=1024-diff;
			}
		} else { /* File at the end */
			if (towrite<1024-diff) { /* Block end into this DES frame */
				int kk,error;
				if (inode->u.tcfs_i.tcfs_fl.bf.spure>0)
					result-=8-inode->u.tcfs_i.tcfs_fl.bf.spure;
/*				printk("Caso 3 ... subcase 1 %ld %ld %d\n",(long)newoff,(long)diff,towrite);*/
				memcpy_fromfs(&mbuf[diff],tmp,towrite);
				kk=(towrite+diff>result)?towrite+diff:result;
				mkendes(mbuf,DES8(kk),ks);
				resw=tcfs_proc_kwrite(inode,newoff,DES8(kk),mbuf,&fattr);
				if (resw<0) {
					kfree(mbuf);
					return resw;
				}
				tmp+=towrite;
				file->f_pos+=towrite;
				inode->u.tcfs_i.tcfs_fl.bf.spure=(int)kk%8;
				inode->i_size = file->f_pos;
				towrite=0;
				error=tcfs_proc_seteattr(NULL,&(inode->i_sb->u.tcfs_sb.x_server),inode->u.tcfs_i.pathname,inode->u.tcfs_i.tcfs_fl.cflag,NULL);
				if (error<0) {
					kfree(mbuf);
					return error;
				}
			} else {
/*				printk("Caso 3 ... subcase 2 %ld\n",(long)newoff);*/
				memcpy_fromfs(&mbuf[diff],tmp,1024-diff);
				mkendes(mbuf,1024,ks);
				resw=tcfs_proc_kwrite(inode,newoff,1024,mbuf,&fattr);
				if (resw<0) {
					kfree(mbuf);
					return resw;
				}
				towrite-=1024-diff;
				tmp+=1024-diff;
				file->f_pos+=1024-diff;
			}
		}
	} while(towrite>0);
	if (inode->i_ino == fattr.fileid) {
		tcfs_refresh_inode(inode, &fattr);
		if (inode->u.tcfs_i.tcfs_fl.bf.spure>0)
			inode->i_size-=8-inode->u.tcfs_i.tcfs_fl.bf.spure;
	}
	kfree(mbuf);
	invalidate_inode_pages(inode);	
	return count;
}

static int tcfs_file_write(struct inode *inode, struct file *file, const char *buf,
			  int count)
{
	int result, written, wsize;
	struct tcfs_fattr fattr;
	unsigned long pos;
	struct hash_entry *htmp;
	struct gid_hash_entry *ghtmp;
	
#ifdef DEBUG_TCFS
	printk("TCFS: tcfs_file_write %s\n",inode->u.tcfs_i.pathname);
#endif	
	TCFS_SEMR_DOWN(inode);

	if (!inode) {
		printk("tcfs_file_write: inode = NULL\n");
		TCFS_SEMR_UP(inode);
		return -EINVAL;
	}
	if (!S_ISREG(inode->i_mode)) {
		printk("tcfs_file_write: write to non-file, mode %07o\n",
			inode->i_mode);
		TCFS_SEMR_UP(inode);
		return -EINVAL;
	}
	if (count <= 0) {
		TCFS_SEMR_UP(inode);
		return 0;
	}

	if (TCFS_IS_SHARED(inode)) {
		ghtmp=gid_hash_lookup(inode->i_gid);
		if (ghtmp==NULL || ghtmp->ks==NULL) {
			TCFS_SEMR_UP(inode);
			return -EACCES;
		}
		result=tcfs_crywrite(inode,file,buf,count,ghtmp->ks);
		TCFS_SEMR_UP(inode);
		return result;
	} 
	if (TCFS_IS_SECURE(inode->u.tcfs_i.tcfs_fl.cflag)) {
		htmp=hash_lookup(current->uid);
		if (htmp==NULL) {
			TCFS_SEMR_UP(inode);
			return -EACCES;
		}
		result=tcfs_crywrite(inode,file,buf,count,htmp->ks);
		TCFS_SEMR_UP(inode);
		return result;
	}
	pos = file->f_pos;
	if (file->f_flags & O_APPEND)
		pos = inode->i_size;
	wsize = TCFS_SERVER(inode)->wsize;
	result = 0;
	written = 0;
	while (written < count) {
		int hunk = count - written;
		if (hunk >= wsize)
			hunk = wsize;
		result = tcfs_proc_write(inode, pos, hunk, buf, &fattr);
		if (result < 0)
			break;
		pos += hunk;
		buf += hunk;
		written += hunk;
		if (hunk < wsize)
			break;
	}
	if (!written) {
		TCFS_SEMR_UP(inode);
		return result;
	}
	file->f_pos = pos;
	if (pos > inode->i_size)
		inode->i_size = pos;
	/* Avoid possible Solaris 2.5 tcfsd bug */
	if (inode->i_ino == fattr.fileid) {
		tcfs_refresh_inode(inode, &fattr);
	}
	TCFS_SEMR_UP(inode);
	return written;
}

==> tcfs-2.2.2.tar.gz: linux/fs/tcfs/file.c <==
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/stat.h>
#include <linux/mm.h>
#include <linux/tcfs_fs.h>
#include <linux/malloc.h>
#include <linux/pagemap.h>

#include <asm/segment.h>
#include <asm/system.h>

#include <linux/tcfs/cipherinterface.h>
#include <linux/tcfs/hash.h>

static int tcfs_file_mmap(struct inode *, struct file *, struct vm_area_struct *);
static int tcfs_file_read(struct inode *, struct file *, char *, int);
static int tcfs_file_write(struct inode *, struct file *, const char *, int);
static int tcfs_fsync(struct inode *, struct file *);

static struct file_operations tcfs_file_operations = {
	NULL,			/* lseek - default */
	tcfs_file_read,		/* read */
	tcfs_file_write,	/* write */
	NULL,			/* readdir - bad */
	NULL,			/* select - default */
	tcfs_ioctl,		/* ioctl - default */
	tcfs_file_mmap,		/* mmap */
	NULL,			/* no special open is needed */
	NULL,			/* release */
	tcfs_fsync,		/* fsync */
};

struct inode_operations tcfs_file_inode_operations = {
	&tcfs_file_operations,	/* default file operations */
	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	tcfs_readpage,		/* readpage */
	NULL,			/* writepage */
	NULL,			/* bmap */
	NULL			/* truncate */
};

static inline void revalidate_inode(struct tcfs_server * server, struct inode * inode)
{
	struct tcfs_fattr fattr;

	if (jiffies - TCFS_READTIME(inode) < TCFS_ATTRTIMEO(inode)) {
		return;
	}
	TCFS_READTIME(inode) = jiffies;
	if (tcfs_proc_getattr(server, TCFS_FH(inode), &fattr) == 0) {
		tcfs_refresh_inode(inode, &fattr);
		if (fattr.mtime.seconds == TCFS_OLDMTIME(inode)) {
			if ((TCFS_ATTRTIMEO(inode) <<= 1) > server->acregmax)
				TCFS_ATTRTIMEO(inode) = server->acregmax;
			if (S_ISREG(inode->i_mode) && 
				(TCFS_IS_SHARED(inode) || 
				TCFS_IS_SECURE(inode->u.tcfs_i.tcfs_fl.cflag)) && 
				inode->i_size>0 && inode->u.tcfs_i.tcfs_fl.bf.spure>0) {
				inode->i_size-=8-inode->u.tcfs_i.tcfs_fl.bf.spure;
			}
			return;
		}
		TCFS_OLDMTIME(inode) = fattr.mtime.seconds;
	}
	if (tcfs_proc_geteattr(&(inode->i_sb->u.tcfs_sb.x_server),inode->u.tcfs_i.pathname,&(inode->u.tcfs_i.tcfs_fl.cflag))<0) {
#ifdef DEBUG_TCFS
	printk("TCFS: tcfs_proc_geteattr returned error\n");
#endif
		inode->u.tcfs_i.tcfs_fl.bf.mb=0;
		inode->u.tcfs_i.tcfs_fl.bf.spure=0;
		inode->u.tcfs_i.tcfs_fl.bf.unknow=0;
	}
	if (S_ISREG(inode->i_mode) && 
		(TCFS_IS_SHARED(inode) || 
		TCFS_IS_SECURE(inode->u.tcfs_i.tcfs_fl.cflag)) && 
		inode->i_size>0 && inode->u.tcfs_i.tcfs_fl.bf.spure>0) {
		inode->i_size-=8-inode->u.tcfs_i.tcfs_fl.bf.spure;
	}
	invalidate_inode_pages(inode);
}

static int tcfs_file_read(struct inode * inode, struct file * file,
	char * buf, int count)
{
	int result;
	unsigned long old;
	struct hash_entry *htmp;
	struct gid_hash_entry *ghtmp;

#ifdef DEBUG_TCFS
	printk("TCFS: tcfs_file_read %s %d\n",inode->u.tcfs_i.pathname,(int)inode->i_size);
#endif
	if (TCFS_IS_SHARED(inode)) {
		ghtmp=gid_hash_lookup(inode->i_gid);
		if (ghtmp==NULL || ghtmp->ks==NULL) {
			return -EACCES;
		}
	} 

	if (TCFS_IS_SECURE(inode->u.tcfs_i.tcfs_fl.cflag)) {
		htmp=hash_lookup(current->uid);
		if (htmp==NULL || current->uid != inode->i_uid) {
			return -EACCES;
		}
	}
	TCFS_SEMR_DOWN(inode);
	/* DES acts in readpage ... here we do a  normal read ...  */
	revalidate_inode(TCFS_SERVER(inode), inode);
	if (inode==NULL || file==NULL || buf==NULL) {
		TCFS_SEMR_UP(inode);
		return -EIO;
	}
	old=file->f_pos;
	result=generic_file_read(inode, file, buf, count);
/*	if (result>0 && TCFS_IS_SECURE(inode->u.tcfs_i.tcfs_fl.cflag) && 
		inode->u.tcfs_i.tcfs_fl.bf.spure>0 && 
		old+result>inode->i_size - 8+inode->u.tcfs_i.tcfs_fl.bf.spure ) {
		result-=8-inode->u.tcfs_i.tcfs_fl.bf.spure;
	}*/
	TCFS_SEMR_UP(inode);
	return result;
}

static int tcfs_file_mmap(struct inode * inode, struct file * file, struct vm_area_struct * vma)
{
	int ret;
	struct hash_entry *htmp;
	struct gid_hash_entry *ghtmp;

#ifdef DEBUG_TCFS
	printk("TCFS: tcfs_file_mmap %s\n",inode->u.tcfs_i.pathname);
#endif	
	if (TCFS_IS_SHARED(inode)) {
		ghtmp=gid_hash_lookup(inode->i_gid);
		if (ghtmp==NULL || ghtmp->ks==NULL) {
			return -EACCES;
		}
	}
	if (TCFS_IS_SECURE(inode->u.tcfs_i.tcfs_fl.cflag)) {
		htmp=hash_lookup(current->uid);
		if (htmp==NULL) {
			return -EACCES;
		}
	}
	TCFS_SEMR_DOWN(inode);
	revalidate_inode(TCFS_SERVER(inode), inode);
	ret=generic_file_mmap(inode, file, vma);
	TCFS_SEMR_UP(inode);
	return ret;
}

static int tcfs_fsync(struct inode *inode, struct file *file)
{
#ifdef DEBUG_TCFS
	printk("TCFS: tcfs_fsync\n");
#endif
	return 0;
}

extern int tcfs_proc_kwrite(struct inode * inode, int offset,int count, const char *data, struct tcfs_fattr *fattr);

static int tcfs_crywrite(struct inode *inode, struct file *file, const char *buf,
	int count, void *ks)
{
	loff_t newoff,diff;
	struct tcfs_fattr fattr;
	int result,resw;
	char *tmp=buf;
	int towrite=count;
	unsigned char *mbuf=NULL;
	
	while(mbuf==NULL) {
		mbuf=kmalloc(1024,GFP_KERNEL);
		schedule();
	}

	if (file->f_flags & O_APPEND) {
		file->f_pos = inode->i_size;
	}
	do {
		schedule();
		newoff=file->f_pos & 0xfffffc00; /* Calculate offset of DES block */
		diff=file->f_pos-newoff;
		memset(mbuf,'\0',1024);
		result=tcfs_proc_read(&(inode->i_sb->u.tcfs_sb.s_server),TCFS_FH(inode),
					newoff,1024,mbuf,&fattr); /* Read old block */
		if (result <0 ) {/* Error reading file */
			kfree(mbuf);
			return result;
		}
		if (result > 0)
			tcfs_decrypt(mbuf,DES8(result),ks);
		if (result==0) { /* End of File */
			if (towrite<1024-diff) { /* Block end into this DES frame */
				int error;
/*				printk("Caso 1 ... subcase 1 %ld\n",(long)newoff);*/
				memcpy_fromfs(&mbuf[diff],tmp,towrite);
				tcfs_encrypt(mbuf,DES8(towrite+diff),ks);
				resw=tcfs_proc_kwrite(inode,newoff,DES8(towrite+diff),mbuf,&fattr);
				if (resw<0) {
					kfree(mbuf);
					return resw;
				}
				file->f_pos+=towrite;
				tmp+=towrite;
				inode->u.tcfs_i.tcfs_fl.bf.spure=(int)(towrite+diff)%8;
				inode->i_size += towrite;
				towrite=0;
				error=tcfs_proc_seteattr(NULL,&(inode->i_sb->u.tcfs_sb.x_server),inode->u.tcfs_i.pathname,inode->u.tcfs_i.tcfs_fl.cflag,NULL);
				if (error<0) {
					kfree(mbuf);
					return error;
				}
			} else { /* Need another DES frame */
/*				printk("Caso 1 ... subcase 2 %ld\n",(long)newoff);*/
				memcpy_fromfs(&mbuf[diff],tmp,1024);
				tcfs_encrypt(mbuf,1024,ks);
				resw=tcfs_proc_kwrite(inode,newoff,1024,mbuf,&fattr);
				if (resw<0) {
					kfree(mbuf);
					return resw;
				}
				file->f_pos+=1024-diff;
				towrite-=1024-diff;
				tmp+=1024-diff;
			}
			continue;
		}
		if (result == 1024) { /* File not at the end */
			if (towrite<1024-diff) { /* Block end into this DES frame */
/*				printk("Caso 2 ... subcase 1 %ld\n",(long)newoff);*/
				memcpy_fromfs(&mbuf[diff],tmp,towrite);
				tcfs_encrypt(mbuf,1024,ks);
				resw=tcfs_proc_kwrite(inode,newoff,1024,mbuf,&fattr);
				if (resw<0) {
					kfree(mbuf);
					return resw;
				}
				tmp+=towrite;
				file->f_pos+=towrite;
				towrite=0;
			} else {
/*				printk("Caso 2 ... subcase 2 %ld\n",(long)newoff);*/
				memcpy_fromfs(&mbuf[diff],tmp,1024-diff);
				tcfs_encrypt(mbuf,1024,ks);
				resw=tcfs_proc_kwrite(inode,newoff,1024,mbuf,&fattr);
				if (resw<0) {
					kfree(mbuf);
					return resw;
				}
				towrite-=1024-diff;
				tmp+=1024-diff;
				file->f_pos+=1024-diff;
			}
		} else { /* File at the end */
			if (towrite<1024-diff) { /* Block end into this DES frame */
				int kk,error;
				if (inode->u.tcfs_i.tcfs_fl.bf.spure>0)
					result-=8-inode->u.tcfs_i.tcfs_fl.bf.spure;
/*				printk("Caso 3 ... subcase 1 %ld %ld %d\n",(long)newoff,(long)diff,towrite);*/
				memcpy_fromfs(&mbuf[diff],tmp,towrite);
				kk=(towrite+diff>result)?towrite+diff:result;
				tcfs_encrypt(mbuf,DES8(kk),ks);
				resw=tcfs_proc_kwrite(inode,newoff,DES8(kk),mbuf,&fattr);
				if (resw<0) {
					kfree(mbuf);
					return resw;
				}
				tmp+=towrite;
				file->f_pos+=towrite;
				inode->u.tcfs_i.tcfs_fl.bf.spure=(int)kk%8;
				inode->i_size = file->f_pos;
				towrite=0;
				error=tcfs_proc_seteattr(NULL,&(inode->i_sb->u.tcfs_sb.x_server),inode->u.tcfs_i.pathname,inode->u.tcfs_i.tcfs_fl.cflag,NULL);
				if (error<0) {
					kfree(mbuf);
					return error;
				}
			} else {
/*				printk("Caso 3 ... subcase 2 %ld\n",(long)newoff);*/
				memcpy_fromfs(&mbuf[diff],tmp,1024-diff);
				tcfs_encrypt(mbuf,1024,ks);
				resw=tcfs_proc_kwrite(inode,newoff,1024,mbuf,&fattr);
				if (resw<0) {
					kfree(mbuf);
					return resw;
				}
				towrite-=1024-diff;
				tmp+=1024-diff;
				file->f_pos+=1024-diff;
			}
		}
	} while(towrite>0);
	if (inode->i_ino == fattr.fileid) {
		tcfs_refresh_inode(inode, &fattr);
		if (inode->u.tcfs_i.tcfs_fl.bf.spure>0)
			inode->i_size-=8-inode->u.tcfs_i.tcfs_fl.bf.spure;
	}
	kfree(mbuf);
	invalidate_inode_pages(inode);	
	return count;
}

static int tcfs_file_write(struct inode *inode, struct file *file, const char *buf,
			  int count)
{
	int result, written, wsize;
	struct tcfs_fattr fattr;
	unsigned long pos;
	struct hash_entry *htmp;
	struct gid_hash_entry *ghtmp;
	
#ifdef DEBUG_TCFS
	printk("TCFS: tcfs_file_write %s\n",inode->u.tcfs_i.pathname);
#endif	
	TCFS_SEMR_DOWN(inode);

	if (!inode) {
		printk("tcfs_file_write: inode = NULL\n");
		TCFS_SEMR_UP(inode);
		return -EINVAL;
	}
	if (!S_ISREG(inode->i_mode)) {
		printk("tcfs_file_write: write to non-file, mode %07o\n",
			inode->i_mode);
		TCFS_SEMR_UP(inode);
		return -EINVAL;
	}
	if (count <= 0) {
		TCFS_SEMR_UP(inode);
		return 0;
	}

	if (TCFS_IS_SHARED(inode)) {
		ghtmp=gid_hash_lookup(inode->i_gid);
		if (ghtmp==NULL || ghtmp->ks==NULL) {
			TCFS_SEMR_UP(inode);
			return -EACCES;
		}
		result=tcfs_crywrite(inode,file,buf,count,ghtmp->ks);
		TCFS_SEMR_UP(inode);
		return result;
	} 
	if (TCFS_IS_SECURE(inode->u.tcfs_i.tcfs_fl.cflag)) {
		htmp=hash_lookup(current->uid);
		if (htmp==NULL) {
			TCFS_SEMR_UP(inode);
			return -EACCES;
		}
		result=tcfs_crywrite(inode,file,buf,count,htmp->ks);
		TCFS_SEMR_UP(inode);
		return result;
	}
	pos = file->f_pos;
	if (file->f_flags & O_APPEND)
		pos = inode->i_size;
	wsize = TCFS_SERVER(inode)->wsize;
	result = 0;
	written = 0;
	while (written < count) {
		int hunk = count - written;
		if (hunk >= wsize)
			hunk = wsize;
		result = tcfs_proc_write(inode, pos, hunk, buf, &fattr);
		if (result < 0)
			break;
		pos += hunk;
		buf += hunk;
		written += hunk;
		if (hunk < wsize)
			break;
	}
	if (!written) {
		TCFS_SEMR_UP(inode);
		return result;
	}
	file->f_pos = pos;
	if (pos > inode->i_size)
		inode->i_size = pos;
	/* Avoid possible Solaris 2.5 tcfsd bug */
	if (inode->i_ino == fattr.fileid) {
		tcfs_refresh_inode(inode, &fattr);
	}
	TCFS_SEMR_UP(inode);
	return written;
}
